// Copyright (c) 2013, Lukas Renggli <renggli@gmail.com>

library bit_set_test;

import 'dart:math';
import 'package:unittest/unittest.dart';
import 'package:more/bit_set.dart';

List<bool> randomList(int seed, int length) {
  var list = new List();
  var generator = new Random(seed);
  for (var i = 0; i < length; i++) {
    list.add(generator.nextBool());
  }
  return list;
}

void main() {
  group('bit list', () {
    test('empty list', () {
      var list = new BitSet(0);
      expect(list, isEmpty);
      expect(list, []);
    });
    test('tiny list', () {
      var list = new BitSet(1);
      expect(list, isNot(isEmpty));
      expect(list, hasLength(1));
      expect(list, [false]);
      list[0] = true;
      expect(list, [true]);
      list[0] = false;
      expect(list, [false]);
    });
    test('long list', () {
      var target = new BitSet(100);
      expect(target, hasLength(100));
      for (var i = 0; i < 100; i++) {
        var source = randomList(311 * i, target.length);
        for (var j = 0; j < source.length; j++) {
          target[j] = source[j];
        }
        for (var j = 0; j < source.length; j++) {
          expect(source[j], target[j]);
        }
      }
    });
    test('list conversion', () {
      for (var len = 0; len < 100; len++) {
        var source = randomList(457 * len, len);
        var target = new BitSet.fromList(source);
        expect(source, target);
        expect(source, target.toList());
        expect(source, new BitSet.fromList(target));
      }
    });
    test('bit counting', () {
      for (var len = 0; len < 100; len++) {
        var list = new BitSet.fromList(randomList(823 * len, len));
        expect(list.count(true) + list.count(false), list.length);
      }
    });
    test('reading bounds', () {
      for (var i = 0; i < 100; i++) {
        var list = new BitSet(i);
        expect(() => list[-1], throwsRangeError);
        for (var j = 0; j < i; j++) {
          expect(() => list[j], returnsNormally);
        }
        expect(() => list[i], throwsRangeError);
      }
    });
    test('writing bounds', () {
      for (var i = 0; i < 100; i++) {
        var list = new BitSet(i);
        expect(() => list[-1] = true, throwsRangeError);
        for (var j = 0; j < i; j++) {
          expect(() => list[j] = true, returnsNormally);
        }
        expect(() => list[i] = true, throwsRangeError);
      }
    });
    test('logical negation', () {
      var source = new BitSet.fromList(randomList(702, 100));
      var target = ~source;
      for (var i = 0; i < target.length; i++) {
        expect(target[i], !source[i]);
      }
    });
    test('logical and', () {
      var source1 = new BitSet.fromList(randomList(439, 100));
      var source2 = new BitSet.fromList(randomList(902, 100));
      var target = source1 & source2;
      for (var i = 0; i < target.length; i++) {
        expect(target[i], source1[i] && source2[i]);
      }
      expect(target, source2 & source1);
    });
    test('logical or', () {
      var source1 = new BitSet.fromList(randomList(817, 100));
      var source2 = new BitSet.fromList(randomList(858, 100));
      var target = source1 | source2;
      for (var i = 0; i < target.length; i++) {
        expect(target[i], source1[i] || source2[i]);
      }
      expect(target, source2 | source1);
    });
    test('fixed length', () {
      var list = new BitSet(32);
      expect(() => list.add(false), throwsUnsupportedError);
      expect(() => list.addAll([true, false]), throwsUnsupportedError);
      expect(() => list.clear(), throwsUnsupportedError);
      expect(() => list.insert(2, true), throwsUnsupportedError);
      expect(() => list.insertAll(2, [true, false]), throwsUnsupportedError);
      expect(() => list.length = 10, throwsUnsupportedError);
      expect(() => list.remove(true), throwsUnsupportedError);
      expect(() => list.removeAll([true, false]), throwsUnsupportedError);
      expect(() => list.removeAt(2), throwsUnsupportedError);
      expect(() => list.removeLast(), throwsUnsupportedError);
      expect(() => list.removeRange(2, 4), throwsUnsupportedError);
      expect(() => list.removeWhere((value) => true), throwsUnsupportedError);
      expect(() => list.replaceRange(2, 4, [true, false]), throwsUnsupportedError);
      expect(() => list.retainAll([true, false]), throwsUnsupportedError);
      expect(() => list.retainWhere((value) => false), throwsUnsupportedError);
    });
  });
}